"""Utility functions for IVOA clients."""

from pyvo.dal import SIA2Service
from pyvo.dal.adhoc import DatalinkResults
from pyvo.dal.sia2 import ObsCoreRecord

from .utils import get_pyvo_auth, get_service_url


def get_datalink_result(result: ObsCoreRecord) -> DatalinkResults:
    """Return the datalink part of a result."""
    return DatalinkResults.from_result_url(
        result.getdataurl(), session=get_pyvo_auth()
    )


def get_siav2_service(label: str) -> SIA2Service:
    """Construct an `SIA2Service` client."""
    if label != "staff":
        raise ValueError(label + " data not available at your location")

    # No matter what, we've only got one sia server per environment
    # so for now just do some checking.
    return SIA2Service(get_service_url("siav2"), get_pyvo_auth())



"""Pytest configuration and fixtures."""

import os
from collections.abc import Iterator
from pathlib import Path
from shutil import copytree
from tempfile import TemporaryDirectory
from unittest.mock import patch

import pytest

from lsst.rsp.startup.storage.command import Command


@pytest.fixture
def _rsp_paths(monkeypatch: pytest.MonkeyPatch) -> Iterator[None]:
    # For each of these, we want to cover both the "from ..constants import"
    # and the "import lsst.rsp.constants" case.
    with patch(
        "lsst.rsp.startup.services.labrunner.TOP_DIR_PATH",
        (Path(__file__).parent / "support" / "files" / "stack_top"),
    ):
        with patch(
            "lsst.rsp.startup.constants.TOP_DIR_PATH",
            (Path(__file__).parent / "support" / "files" / "stack_top"),
        ):
            with patch(
                "lsst.rsp.startup.services.labrunner.ETC_PATH",
                (Path(__file__).parent / "support" / "files" / "etc"),
            ):
                with patch(
                    "lsst.rsp.startup.constants.ETC_PATH",
                    (Path(__file__).parent / "support" / "files" / "etc"),
                ):
                    yield


@pytest.fixture
def _rsp_env(
    _rsp_paths: None, monkeypatch: pytest.MonkeyPatch
) -> Iterator[None]:
    template = Path(__file__).parent / "support" / "files" / "homedir"
    with TemporaryDirectory() as homedir:
        monkeypatch.setenv("HOME", homedir)
        monkeypatch.setenv("USER", "hambone")
        monkeypatch.setenv("JUPYTERHUB_BASE_URL", "/nb/")
        copytree(
            template,
            homedir,
            dirs_exist_ok=True,
            symlinks=True,
        )
        yield


@pytest.fixture
def git_repo() -> Iterator[Path]:
    with TemporaryDirectory() as repo_str:
        pwd = Path.cwd()
        repo = Path(repo_str)
        os.chdir(repo)
        cmd = Command()
        cmd.run("git", "init")
        (repo / "README.md").write_text("# Test Repo\n")
        cmd.run("git", "config", "user.email", "hambone@opera.borphee.quendor")
        cmd.run("git", "config", "user.name", "Hambone")
        cmd.run("git", "config", "init.defaultBranch", "main")
        cmd.run("git", "checkout", "-b", "main")
        cmd.run("git", "add", "README.md")
        cmd.run("git", "commit", "-am", "Initial Commit")
        os.chdir(pwd)
        yield Path(repo)



"""Set up custom logging for RSP.

At startup, if ${HOME}/.ipython/profile_default/startup/20-logging.py does
not exist, this file will be copied to it from /opt/lsst/software/jupyterlab.

It will also be copied if that file exists but is an earlier standard version
(as determined via sha256sum).

If you don't like what it does, create an empty file (or one that does what
you want) at ${HOME}/.ipython/profile_default/startup/20-logging.py and it
will not be recopied.
"""

import logging
import os
import sys

customlogger = False

try:
    from lsst.rsp import IPythonHandler, forward_lsst_log

    customlogger = True
except ImportError:
    pass  # Probably a container that doesn't have our new code

# If the whole container is in debug mode, enable debug logging by default.
# Otherwise, use the default level (which is warning); however, lsst logs
# should be at info level.
debug = os.getenv("DEBUG")
handlers = []
if customlogger:
    # Forward anything at INFO or above, unless debug is set, in which case,
    # forward DEBUG and above.
    if debug:
        forward_lsst_log("DEBUG")
    else:
        forward_lsst_log("INFO")
    handlers = [IPythonHandler()]
else:
    # Set up WARNING and above as stderr, below that to stdout.  This is
    # intended to make GKE error reporting more consistent and to correspond
    # to the usual Unix distinction between error and non-error output.
    warnhandler = logging.StreamHandler(stream=sys.stderr)
    warnhandler.setLevel(logging.WARNING)
    handlers = [warnhandler]
    if debug:
        lowhandler = logging.StreamHandler(stream=sys.stdout)
        lowhandler.setLevel(logging.DEBUG)
        handlers.append(lowhandler)
logging.basicConfig(force=True, handlers=handlers)
# Now set up INFO for lsst logs everywhere
logging.getLogger("lsst").setLevel(logging.INFO)



"""Test the packaging."""

from __future__ import annotations

from lsst.rsp import __version__


def test_version() -> None:
    """Ensure that the version is set."""
    assert isinstance(__version__, str)
    # Indicates the package is not installed otherwise
    assert __version__ != "0.0.0"



"""Constants for RSP startup."""

from pathlib import Path

__all__ = [
    "APP_NAME",
    "ETC_PATH",
    "PREVIOUS_LOGGING_CHECKSUMS",
    "MAX_NUMBER_OUTPUTS",
    "NONINTERACTIVE_CONFIG_PATH",
    "TOP_DIR_PATH",
]

APP_NAME = "nublado"
"""Application name, used for logging."""

ETC_PATH = Path("/etc")
"""Configuration directory, usually /etc, but overrideable for tests."""

PREVIOUS_LOGGING_CHECKSUMS = [
    "2997fe99eb12846a1b724f0b82b9e5e6acbd1d4c29ceb9c9ae8f1ef5503892ec"
]
"""sha256 sums of previous iterations of ``20-logging.py``.

Used to determine whether upgrading the logging configuration is
needed, or whether the user has made local modifications that
therefore should not be touched.
"""

MAX_NUMBER_OUTPUTS = 10000
"""Maximum number of output lines to display in a Jupyter notebook cell.

Used to prevent OOM-killing if some cell generates a lot of output.
"""

TOP_DIR_PATH = Path("/opt/lsst/software")
"""
Location where the DM stack and our Lab machinery are rooted.

Overrideable for testing.
"""

NONINTERACTIVE_CONFIG_PATH = Path(
    TOP_DIR_PATH / "jupyterlab" / "noninteractive" / "command" / "command.json"
)
"""
Location where a noninteractive pod will mount its command configuration.
"""



"""Collection of utilities for Rubin Science Platform notebooks."""

from importlib.metadata import PackageNotFoundError, version

from .catalog import (
    get_catalog,
    get_obstap_service,
    get_tap_service,
    retrieve_query,
)
from .client import RSPClient
from .log import IPythonHandler, forward_lsst_log
from .service import get_datalink_result, get_siav2_service
from .utils import (
    format_bytes,
    get_access_token,
    get_digest,
    get_hostname,
    get_node,
    get_pod,
)

__version__: str
"""The application version string of (PEP 440 / SemVer compatible)."""

try:
    __version__ = version(__name__)
except PackageNotFoundError:
    # package is not installed
    __version__ = "0.0.0"


__all__ = [
    "__version__",
    "IPythonHandler",
    "RSPClient",
    "format_bytes",
    "forward_lsst_log",
    "get_access_token",
    "get_catalog",
    "get_datalink_result",
    "get_digest",
    "get_node",
    "get_pod",
    "get_tap_service",
    "get_siav2_service",
    "get_obstap_service",
    "retrieve_query",
    "get_hostname",
]



"""Client for other services running in the same RSP instance."""

from pathlib import Path

import httpx


class RSPClient(httpx.AsyncClient):
    """Configured client for other services in the RSP.

    It uses knowledge present in the Lab instance it's running in to
    configure token authentication and a base URL.
    """

    def __init__(
        self,
        service_endpoint: str,
        *,
        jupyterlab_dir: Path = Path("/opt/lsst/software/jupyterlab"),
    ) -> None:
        token = (jupyterlab_dir / "secrets" / "token").read_text().strip()
        instance_url = (
            (jupyterlab_dir / "environment" / "EXTERNAL_INSTANCE_URL")
            .read_text()
            .strip()
        )
        if instance_url.endswith("/") or service_endpoint.startswith("/"):
            service_root = f"{instance_url}{service_endpoint}"
        else:
            service_root = f"{instance_url}/{service_endpoint}"
        http_headers = {
            "Authorization": f"Bearer {token}",
            "Content-Type": "application/json",
        }
        super().__init__(
            base_url=service_root, follow_redirects=True, headers=http_headers
        )



"""Test the RSPClient."""

from pathlib import Path

import pytest
from pytest_httpx import HTTPXMock

from lsst.rsp import RSPClient


@pytest.mark.usefixtures("_rsp_env")
@pytest.mark.asyncio
async def test_client(httpx_mock: HTTPXMock) -> None:
    """Ensure that the RSPClient has the right headers and assembles its
    URL correctly.
    """
    httpx_mock.add_response(
        url="https://rsp.example.com/test-service/foo",
        match_headers={
            "Authorization": "Bearer gf-dummytoken",
            "Content-Type": "application/json",
        },
    )
    jupyterlab_dir = (
        Path(__file__).parent / "support" / "files" / "client" / "jupyterlab"
    )
    client = RSPClient("/test-service", jupyterlab_dir=jupyterlab_dir)
    await client.get("/foo")
    # The httpx mock will throw an error at teardown if we did not exercise
    # the mock, so we know the request matched both the URL and the headers.



"""Launcher for the Lab Runner."""

from .services.labrunner import LabRunner


def main() -> None:
    """Make a LabRunner and call its single public method.  All settings are
    in the environment.
    """
    LabRunner().go()
